/****************************************************************************
** $Id: qt/qwaitcondition_win.cpp   3.3.2   edited Aug 13 2003 $
**
** Implementation of QWaitCondition class for Win32
**
** Copyright (C) 2000-2001 Trolltech AS.  All rights reserved.
**
** This file is part of the tools module of the Qt GUI Toolkit.
**
** Licensees holding valid Qt Enterprise Edition or Qt Professional Edition
** licenses for Windows may use this file in accordance with the Qt Commercial
** License Agreement provided with the Software.
**
** This file is not available for use under any other license without
** express written permission from the copyright holder.
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
** See http://www.trolltech.com/pricing.html or email sales@trolltech.com for
**   information about Qt Commercial License Agreements.
**
** Contact info@trolltech.com if any conditions of this licensing are
** not clear to you.
**
**********************************************************************/

#if defined(QT_THREAD_SUPPORT)

#include "qwaitcondition.h"
#include "qnamespace.h"
#include "qmutex.h"
#include "qptrlist.h"
#include "qt_windows.h"

#define Q_MUTEX_T void*
#include <private/qmutex_p.h>
#include <private/qcriticalsection_p.h>

//***********************************************************************
// QWaitConditionPrivate
// **********************************************************************

class QWaitConditionEvent
{
public:
    inline QWaitConditionEvent() : priority(0)
    {
	QT_WA ({
	    event = CreateEvent(NULL, TRUE, FALSE, NULL);
	}, {
	    event = CreateEventA(NULL, TRUE, FALSE, NULL);
	});
    }
    inline ~QWaitConditionEvent() { CloseHandle(event); }
    int priority;
    HANDLE event;
};

typedef QPtrList<QWaitConditionEvent> EventQueue;

class QWaitConditionPrivate
{
public:
    QCriticalSection cs;
    EventQueue queue;
    EventQueue freeQueue;

    bool wait(QMutex *mutex, unsigned long time);
};

bool QWaitConditionPrivate::wait(QMutex *mutex, unsigned long time)
{
    bool ret = FALSE;

    cs.enter();
    QWaitConditionEvent *wce = freeQueue.take();
    if (!wce)
	wce = new QWaitConditionEvent;
    wce->priority = GetThreadPriority(GetCurrentThread());

    // insert 'wce' into the queue (sorted by priority)
    QWaitConditionEvent *current = queue.first();
    int index = 0;
    while (current && current->priority >= wce->priority) {
	current = queue.next();
	++index;
    }
    queue.insert(index, wce);
    cs.leave();

    if (mutex) mutex->unlock();

    // wait for the event
    switch (WaitForSingleObject(wce->event, time)) {
    default: break;

    case WAIT_OBJECT_0:
	ret = TRUE;
	break;
    }

    if (mutex) mutex->lock();

    cs.enter();
    // remove 'wce' from the queue
    queue.removeRef(wce);
    ResetEvent(wce->event);
    freeQueue.append(wce);
    cs.leave();

    return ret;
}

//***********************************************************************
// QWaitCondition implementation
//***********************************************************************

QWaitCondition::QWaitCondition()
{
    d = new QWaitConditionPrivate;
    d->freeQueue.setAutoDelete(TRUE);
}

QWaitCondition::~QWaitCondition()
{ 
    Q_ASSERT(d->queue.isEmpty());
    delete d; 
}

bool QWaitCondition::wait( unsigned long time )
{
    return d->wait(0, time);
}

bool QWaitCondition::wait( QMutex *mutex, unsigned long time)
{
    if ( !mutex )
	return FALSE;

    if ( mutex->d->type() == Q_MUTEX_RECURSIVE ) {
#ifdef QT_CHECK_RANGE
	qWarning("QWaitCondition::wait: Cannot wait on recursive mutexes.");
#endif
	return FALSE;
    }
    return d->wait(mutex, time);
}

void QWaitCondition::wakeOne()
{
    // wake up the first thread in the queue
    d->cs.enter();
    QWaitConditionEvent *first = d->queue.first();
    if (first)
	SetEvent(first->event);
    d->cs.leave();
}

void QWaitCondition::wakeAll()
{
    // wake up the all threads in the queue
    d->cs.enter();
    QWaitConditionEvent *current = d->queue.first();
    while (current) {
	SetEvent(current->event);
	current = d->queue.next();
    }
    d->cs.leave();
}

#endif // QT_THREAD_SUPPORT
